@@ -18,7 +18,6 @@
 #include <stdlib.h>
 #include <string.h>
 #include <glib.h>
-#include <gio/gio.h>
 #include <gst/gst.h>
 #include <espeak/speak_lib.h>
 
@@ -38,7 +37,8 @@ typedef enum
 
 typedef enum
 {
-    INPROCESS = 1
+    INPROCESS = 1,
+    CLOSE     = 2
 } ContextState;
 
 typedef struct
@@ -47,16 +47,17 @@ typedef struct
 
     volatile SpinState state;
 
-    GMemoryOutputStream *sound;
-    goffset sound_offset;
+    GByteArray *sound;
+    gsize sound_offset;
+    GstClockTime audio_position;
 
     GArray *events;
-    goffset events_pos;
+    gsize events_pos;
 
-    goffset last_word;
-    goffset mark_offset;
+    int last_word;
+    int mark_offset;
     const gchar *mark_name;
-    goffset last_mark;
+    int last_mark;
 } Espin;
 
 struct _Econtext
@@ -64,7 +65,7 @@ struct _Econtext
     volatile ContextState state;
 
     gchar *text;
-    goffset text_offset;
+    gsize text_offset;
     gsize text_len;
 
     Espin queue[SPIN_QUEUE_SIZE];
@@ -116,8 +117,17 @@ emit_mark(Econtext *self, guint offset, const gchar *mark)
     gst_bus_post(self->bus, msg);
 }
 
+static inline gsize
+strbstr(Econtext *self, gsize pos, const gchar *needle, gsize needle_len)
+{
+    for (pos -= needle_len; pos >= needle_len; --pos)
+        if (strncmp(self->text + pos, needle, needle_len) == 0)
+            return pos;
+    return 0;
+}
+
 static void init();
-static void process_push(Econtext*);
+static void process_push(Econtext*, gboolean);
 static void process_pop(Econtext*);
 
 static GThread *process_tid = NULL;
@@ -144,8 +154,7 @@ espeak_new(GstElement *emitter)
 
         spin->context = self;
         spin->state = IN;
-        spin->sound = G_MEMORY_OUTPUT_STREAM(g_memory_output_stream_new(
-                    NULL, 0, realloc, free));
+        spin->sound = g_byte_array_new();
         spin->events = g_array_new(FALSE, FALSE, sizeof(espeak_EVENT));
     }
 
@@ -181,9 +190,7 @@ espeak_unref(Econtext *self)
 
     for (i = SPIN_QUEUE_SIZE; i--;)
     {
-        g_output_stream_close(G_OUTPUT_STREAM(self->queue[i].sound),
-                NULL, NULL);
-        g_object_unref(self->queue[i].sound);
+        g_byte_array_free(self->queue[i].sound, TRUE);
         g_array_free(self->queue[i].events, TRUE);
     }
 
@@ -210,7 +217,7 @@ espeak_in(Econtext *self, const gchar *text)
     self->text_offset = 0;
     self->text_len = strlen(text);
 
-    process_push(self);
+    process_push(self, TRUE);
 }
 
 GstBuffer*
@@ -218,63 +225,47 @@ play(Econtext *self, Espin *spin, gsize size_to_play)
 {
     inline gsize whole(Espin *spin, gsize size_to_play)
     {
-        gsize spin_size = g_memory_output_stream_get_data_size(spin->sound);
-        return MIN(size_to_play, spin_size - spin->sound_offset);
+        for (;; ++spin->events_pos)
+        {
+            espeak_EVENT *i = &g_array_index(spin->events, espeak_EVENT,
+                    spin->events_pos);
+            gsize len = i->sample*2 - spin->sound_offset;
+
+            if (i->type == espeakEVENT_LIST_TERMINATED || len >= size_to_play)
+                return len;
+        }
     }
 
     inline gsize word(Econtext *self, Espin *spin, gsize size_to_play)
     {
-        gsize spin_size = g_memory_output_stream_get_data_size(spin->sound);
-        size_to_play = MIN(size_to_play, spin_size);
-
-        goffset event;
-        goffset sample_offset = 0;
-        goffset text_offset = -1;
-        gsize text_len = 0;
+        gsize spin_size = spin->sound->len;
+        gsize event;
+        gsize sample_offset = 0;
 
         for (event = spin->events_pos; TRUE; ++event)
         {
             espeak_EVENT *i = &g_array_index(spin->events, espeak_EVENT, event);
 
-            GST_DEBUG("size_to_play=%ld event=%ld "
-                      "i->type=%d i->text_position=%d",
-                      size_to_play, event, i->type, i->text_position);
+            GST_DEBUG("event=%zd i->type=%d i->text_position=%d",
+                      event, i->type, i->text_position);
 
             if (i->type == espeakEVENT_LIST_TERMINATED)
             {
-                GST_DEBUG("i->sample=%d", i->sample*2);
                 sample_offset = spin_size;
                 break;
             }
             else if (i->type == espeakEVENT_WORD)
             {
+                if (i->text_position != spin->last_word)
+                {
+                    emit_word(self, i->text_position, i->length);
+                    spin->last_word = i->text_position;
+                }
                 sample_offset = i[1].sample*2;
-                text_offset = i->text_position;
-                text_len = i->length;
-
-                GST_DEBUG("sample_offset=%d txt_offset=%d txt_len=%d, txt=%s",
-                        sample_offset, text_offset, text_len,
-                        self->text + text_offset);
                 break;
             }
         }
 
-        if (text_offset != -1 && text_offset > spin->last_word)
-        {
-            spin->last_word = text_offset + text_len;
-            emit_word(self, text_offset, text_len);
-        }
-
-        if (sample_offset - spin->sound_offset > size_to_play)
-        {
-            GST_DEBUG("sample_offset=%ld spin->sound_offset=%ld",
-                    sample_offset, spin->sound_offset);
-            return size_to_play;
-        }
-
-        if (text_offset != -1)
-            spin->events_pos = event + 1;
-
         return sample_offset - spin->sound_offset;
     }
 
@@ -283,15 +274,13 @@ play(Econtext *self, Espin *spin, gsize size_to_play)
         if (spin->mark_name)
         {
             emit_mark(self, spin->mark_offset, spin->mark_name);
-            spin->mark_offset = -1;
+            spin->mark_offset = 0;
             spin->mark_name = NULL;
         }
 
-        gsize spin_size = g_memory_output_stream_get_data_size(spin->sound);
-        size_to_play = MIN(size_to_play, spin_size);
-
-        goffset event;
-        goffset sample_offset = 0;
+        gsize spin_size = spin->sound->len;
+        gsize event;
+        gsize sample_offset = 0;
         guint mark_offset = 0;
         const gchar *mark_name = NULL;
 
@@ -299,9 +288,8 @@ play(Econtext *self, Espin *spin, gsize size_to_play)
         {
             espeak_EVENT *i = &g_array_index(spin->events, espeak_EVENT, event);
 
-            GST_DEBUG("size_to_play=%ld event=%ld "
-                      "i->type=%d i->text_position=%d",
-                      size_to_play, event, i->type, i->text_position);
+            GST_DEBUG("event=%zd i->type=%d i->text_position=%d",
+                      event, i->type, i->text_position);
 
             if (i->type == espeakEVENT_LIST_TERMINATED)
             {
@@ -324,16 +312,8 @@ play(Econtext *self, Espin *spin, gsize size_to_play)
             }
         }
 
-        if (sample_offset - spin->sound_offset > size_to_play)
-        {
-            GST_DEBUG("sample_offset=%ld spin->sound_offset=%ld",
-                    sample_offset, spin->sound_offset);
-            return size_to_play;
-        }
-
         spin->mark_offset = mark_offset;
         spin->mark_name = mark_name;
-        spin->events_pos = event + 1;
 
         return sample_offset - spin->sound_offset;
     }
@@ -352,15 +332,21 @@ play(Econtext *self, Espin *spin, gsize size_to_play)
             size_to_play = whole(spin, size_to_play);
     }
 
+    espeak_EVENT *event = &g_array_index(spin->events, espeak_EVENT,
+            spin->events_pos);
+
     GstBuffer *out = gst_buffer_new();
-    GST_BUFFER_DATA(out) =
-            (guchar*)g_memory_output_stream_get_data(spin->sound) +
-                spin->sound_offset;
+    GST_BUFFER_DATA(out) = spin->sound->data + spin->sound_offset;
     GST_BUFFER_SIZE(out) = size_to_play;
+    GST_BUFFER_TIMESTAMP(out) = spin->audio_position;
+    spin->audio_position = gst_util_uint64_scale_int(
+            event->audio_position, GST_SECOND, 1000);
+    GST_BUFFER_DURATION(out) = spin->audio_position - GST_BUFFER_TIMESTAMP(out);
 
     spin->sound_offset += size_to_play;
+    spin->events_pos += 1;
 
-    GST_DEBUG("out=%p size_to_play=%ld tell=%ld", GST_BUFFER_DATA(out),
+    GST_DEBUG("out=%p size_to_play=%zd tell=%zd", GST_BUFFER_DATA(out),
             size_to_play, spin->sound_offset);
 
     return out;
@@ -374,30 +360,35 @@ espeak_out(Econtext *self, gsize size_to_play)
     for (;;)
     {
         g_mutex_lock(process_lock);
-            while ((g_atomic_int_get(&self->out->state) & (PLAY|OUT)) == 0)
+            for (;;)
             {
-                if ((self->state & INPROCESS) == 0)
+                if (g_atomic_int_get(&self->out->state) & (PLAY|OUT))
+                    break;
+                if (self->state != INPROCESS)
                 {
-                    GST_DEBUG("[%p]", self);
+                    if (self->state == CLOSE)
+                        GST_DEBUG("[%p] sesseion is closed", self);
+                    else
+                        GST_DEBUG("[%p] nothing to play", self);
                     g_mutex_unlock(process_lock);
                     return NULL;
                 }
-                GST_DEBUG("[%p]", self);
+                GST_DEBUG("[%p] wait for processed data", self);
                 g_cond_wait(process_cond, process_lock);
             }
         g_mutex_unlock(process_lock);
 
         Espin *spin = self->out;
-        gsize spin_size = g_memory_output_stream_get_data_size(spin->sound);
+        gsize spin_size = spin->sound->len;
 
-        GST_DEBUG("[%p] spin->sound_offset=%ld spin_size=%ld",
+        GST_DEBUG("[%p] spin->sound_offset=%zd spin_size=%zd",
                 self, spin->sound_offset, spin_size);
 
         if (g_atomic_int_get(&spin->state) == PLAY &&
                 spin->sound_offset >= spin_size)
         {
             g_atomic_int_set(&spin->state, IN);
-            process_push(self);
+            process_push(self, FALSE);
             spinning(self->queue, &self->out);
             continue;
         }
@@ -438,53 +429,61 @@ synth_cb(short *data, int numsamples, espeak_EVENT *events)
         return 0;
 
     Espin *spin = events->user_data;
+    Econtext *self = spin->context;
 
     if (numsamples > 0)
     {
-        g_output_stream_write(G_OUTPUT_STREAM(spin->sound), data, numsamples*2,
-                NULL, NULL);
+        g_byte_array_append(spin->sound, (const guint8*)data, numsamples*2);
 
         espeak_EVENT *i;
 
         for (i = events; i->type != espeakEVENT_LIST_TERMINATED; ++i)
         {
-            if (i->type == espeakEVENT_WORD)
-                --i->text_position;
-            else if (i->type == espeakEVENT_MARK)
-            {
-                // suppress failed text_position values
-                if (spin->last_mark)
-                {
-                    goffset pos = strstr(spin->context->text +
-                            spin->last_mark, "/>") - spin->context->text + 2;
-                    if (i->text_position <= spin->last_mark ||
-                            pos > i->text_position)
-                        i->text_position = pos;
-                }
-
-                spin->last_mark = i->text_position;
-
-                gchar *pos = spin->context->text + i->text_position;
-                gint turn = 0;
-
-                for (; pos > spin->context->text; --pos)
-                    if (*pos == '"')
-                    {
-                        if (turn++ == 0)
-                            *pos = 0;
-                        else
-                        {
-                            i->id.name = pos + 1;
-                            break;
-                        }
-                    }
-            }
-
             GST_DEBUG("type=%d text_position=%d length=%d "
                       "audio_position=%d sample=%d",
                     i->type, i->text_position, i->length,
                     i->audio_position, i->sample*2);
 
+            // convert to 0-based position
+            --i->text_position;
+
+            // workaround to fix text_position related faults for mark events
+            if (i->type == espeakEVENT_MARK)
+            {
+                // suppress failed text_position values
+                const gchar *eom = strstr(self->text +
+                        spin->last_mark, "/>");
+                if (eom)
+                {
+                    int pos = eom - self->text + 2;
+
+                    if (i->text_position <= spin->last_mark ||
+                            pos > i->text_position)
+                        i->text_position = pos;
+                }
+                else if (i->text_position <= spin->last_mark)
+                {
+                    i->text_position = spin->last_mark;
+                }
+
+                spin->last_mark = i->text_position;
+
+                // point mark name to text substring instead of using
+                // espeak's allocated string
+                int l_quote, r_quote;
+                if ((r_quote = strbstr(self, i->text_position, "/>", 2)) != 0)
+                    if ((r_quote = strbstr(self, r_quote, "\"", 1)) != 0)
+                        if ((l_quote = strbstr(self, r_quote, "\"", 1)) != 0)
+                        {
+                            i->id.name = self->text + l_quote + 1;
+                            self->text[r_quote] = 0;
+                        }
+
+            }
+
+            GST_DEBUG("text_position=%d length=%d",
+                    i->text_position, i->length);
+
             g_array_append_val(spin->events, *i);
         }
     }
@@ -497,12 +496,12 @@ synth_cb(short *data, int numsamples, espeak_EVENT *events)
 static void
 synth(Econtext *self, Espin *spin)
 {
-    g_seekable_seek(G_SEEKABLE(spin->sound), 0, G_SEEK_SET,
-            NULL, NULL);
+    g_byte_array_set_size(spin->sound, 0);
     g_array_set_size(spin->events, 0);
     spin->sound_offset = 0;
+    spin->audio_position = 0;
     spin->events_pos = 0;
-    spin->mark_offset = -1;
+    spin->mark_offset = 0;
     spin->mark_name = NULL;
     spin->last_word = -1;
     spin->last_mark = 0;
@@ -518,7 +517,7 @@ synth(Econtext *self, Espin *spin)
     if (track == ESPEAK_TRACK_MARK)
         flags |= espeakSSML;
 
-    GST_DEBUG("[%p] text_offset=%ld", self, self->text_offset);
+    GST_DEBUG("[%p] text_offset=%zd", self, self->text_offset);
 
     espeak_Synth(self->text, self->text_len + 1, 0, POS_CHARACTER, 0, flags,
             NULL, spin);
@@ -530,7 +529,7 @@ synth(Econtext *self, Espin *spin)
     }
 
     espeak_EVENT last_event = { espeakEVENT_LIST_TERMINATED };
-    last_event.sample = g_memory_output_stream_get_data_size(spin->sound) / 2;
+    last_event.sample = spin->sound->len / 2;
     g_array_append_val(spin->events, last_event);
 }
 
@@ -609,9 +608,15 @@ process(gpointer data)
 
             process_queue = g_slist_remove_link(process_queue, process_queue);
 
+            if (context->state == CLOSE)
+            {
+                GST_DEBUG("[%p] session is closed", context);
+                continue;
+            }
+
             if (context->text_offset >= context->text_len)
             {
-                GST_DEBUG("[%p]", context);
+                GST_DEBUG("[%p] end of text to process", context);
                 context->state &= ~INPROCESS;
             }
             else
@@ -622,13 +627,13 @@ process(gpointer data)
 
                 if (g_atomic_int_get(&context->in->state) == IN)
                 {
-                    GST_DEBUG("[%p]", context);
+                    GST_DEBUG("[%p] continue to process data", context);
                     process_queue = g_slist_concat(process_queue,
                             context->process_chunk);
                 }
                 else
                 {
-                    GST_DEBUG("[%p]", context);
+                    GST_DEBUG("[%p] pause to process data", context);
                     context->state &= ~INPROCESS;
                 }
             }
@@ -643,34 +648,36 @@ process(gpointer data)
 }
 
 static void
-process_push(Econtext *context)
+process_push(Econtext *context, gboolean force_in)
 {
-    GST_DEBUG("[%p]", context);
+    GST_DEBUG("[%p] lock", context);
     g_mutex_lock(process_lock);
 
-    if ((context->state & INPROCESS) == 0)
+    if (context->state == CLOSE && !force_in)
+        GST_DEBUG("[%p] state=%d", context, context->state);
+    else if (context->state != INPROCESS)
     {
-        context->state |= INPROCESS;
+        context->state = INPROCESS;
         process_queue = g_slist_concat(process_queue, context->process_chunk);
         g_cond_broadcast(process_cond);
     }
 
     g_mutex_unlock(process_lock);
-    GST_DEBUG("[%p]", context);
+    GST_DEBUG("[%p] unlock", context);
 }
 
 static void
 process_pop(Econtext *context)
 {
-    GST_DEBUG("[%p]", context);
+    GST_DEBUG("[%p] lock", context);
     g_mutex_lock(process_lock);
 
     process_queue = g_slist_remove_link(process_queue, context->process_chunk);
-    context->state &= ~INPROCESS;
+    context->state = CLOSE;
     g_cond_broadcast(process_cond);
 
     g_mutex_unlock(process_lock);
-    GST_DEBUG("[%p]", context);
+    GST_DEBUG("[%p] unlock", context);
 }
 
 // -----------------------------------------------------------------------------
